
//ahb_master
module ahb_master
(
    input               send_en,
    input               rcv_en,
    input               rd_config_en,

    input               hclk,
    input               rstn,
    input       [31:0]  t_hrdata,
    input               t_hready,
    input       [1:0]   t_hresp,
    output      [1:0]   t_htrans,
    output              t_hwrite,
    output              t_hsel,
    output  reg [31:0]  t_hwdata,
    output  reg [31:0]  t_haddr,
    output      [2:0]   t_hsize,
    output              t_hready_in    
);

assign t_htrans = 2'b10;
assign t_hsize = 3'b010;
assign t_hready_in = 1'b1;

reg [2:0] c_state,n_state;

localparam s_idle = 3'd0;
localparam s_send = 3'd1;
localparam s_rcv = 3'd2;

reg [4:0] cnt;

parameter SEND_DATA_addr = 32'h20000100;

wire [31:0] send_data1, send_data2, send_data3, send_data4;

assign send_data1 = 32'h01234567;
assign send_data2 = 32'h89abcdef;
assign send_data3 = 32'hfedcba98;
assign send_data4 = 32'h76543210;

parameter RCV_DATA_addr = 32'h20000100;

//send_const is to control to write/read one of send_data1~4.
parameter send_const = 4;   //1:send_data1; 2:send_data2; 3:send_data3; 4:send_data4
parameter rcv_const = 1;

reg user_rd_vld;
reg [31:0] user_rd_data;
reg ahb_state;

always@(posedge hclk or negedge rstn) begin
    if (!rstn) begin
        c_state <= s_idle;
    end
    else begin
        c_state <= n_state;
    end
end

always@(*) begin
    n_state = c_state;
    case(c_state)
    s_idle:
        if (send_en) n_state = s_send;
        else if (rcv_en) n_state = s_rcv;
    s_send:
        if (cnt == send_const-1 && t_hready && ahb_state) n_state =s_idle;
    s_rcv:
        if (cnt == rcv_const-1 && t_hready && ahb_state) n_state =s_idle;  
    default:
        n_state = s_idle;
    endcase
end

always@(posedge hclk or negedge rstn) begin
    if (!rstn) begin
        ahb_state <= 1'b0;
    end
    else begin
        if (t_hsel & t_hready & !ahb_state) begin
            ahb_state <= 1'b1;
        end
        else if (ahb_state & t_hready) begin
            ahb_state <= 1'b0;
        end
    end
end

assign t_hwrite = ((c_state == s_send) && !ahb_state) & t_hready;
assign t_hsel = (((c_state == s_send) && !ahb_state) || ((c_state == s_rcv) && !ahb_state)) & t_hready;

always@(posedge hclk or negedge rstn) begin
    if (!rstn) begin
        cnt <= 0;
    end
    else begin
        if (c_state == s_idle) begin
            cnt <= 0;
        end
        else begin
            if (ahb_state & t_hready) begin
                cnt <= cnt + 1;
            end
        end
    end
end

always@(*) begin
    if (c_state == s_send) begin
        t_haddr = SEND_DATA_addr;
        if (cnt == 0) begin
            t_hwdata = send_data1;
        end
        else if (cnt == 1) begin
            t_hwdata = send_data2;
        end
        else if (cnt == 2) begin
            t_hwdata = send_data3;
        end
        else if (cnt == 3) begin
            t_hwdata = send_data4;
        end
        else begin
            t_hwdata = 0;
        end
    end
    else if (c_state == s_rcv) begin
        t_haddr = RCV_DATA_addr;
        t_hwdata = 0;
    end
    else begin
        t_haddr = 0;
        t_hwdata = 0;
    end
end

always@(posedge hclk or negedge rstn) begin
    if (!rstn) begin
        user_rd_vld <= 1'b0;
        user_rd_data <= 32'd0;
    end
    else begin
        if (c_state == s_rcv & ahb_state & t_hready) begin
            user_rd_vld <= 1'b1;
            user_rd_data <= t_hrdata;
        end
        else begin
            user_rd_vld <= 1'b0;
        end
    end
end

endmodule