
//apb_fifo
module apb_fifo
(
    input          pclk,       //master clock input
    input          presetn,    //synchronous active low reset    
    input          psel,       //slave sel
    input          penable,    //enable
    input          pwrite,     //transmit direction
    input   [11:2] paddr,      //lower address bits
    input   [31:0] pwdata,     //write data
    output  [31:0] prdata,     //read data
    output         pready,     //ready
    output         pslverr,
    input          write_en,
    input          read_en,
    input   [11:2] user_addr,
    input   [31:0] user_wdata,
    output  [31:0] user_rdata
);

wire write_enable = psel ? psel & pwrite & penable : write_en;
wire read_enable  = psel ? psel & (!pwrite) & (!penable) : read_en;

wire [4:0] Wnum;
wire [4:0] Rnum;

wire Almost_Empty;
wire Almost_Full;

wire Empty;
wire Full;

wire [31:0] wdata = psel ? pwdata : user_wdata;
wire [31:0] rdata;

//fifo_top instantiation
fifo_top u_fifo_top
(
    .Data(wdata),
    .Reset(~presetn),
    .WrClk(pclk),
    .RdClk(pclk),
    .WrEn(write_enable),
    .RdEn(read_enable),
    .Wnum(Wnum),
    .Rnum(Rnum),
    .Almost_Empty(Almost_Empty),
    .Almost_Full(Almost_Full),
    .Q(rdata),
    .Empty(Empty),
    .Full(Full)
);

assign user_rdata = rdata;

assign prdata = rdata;
assign pready = 1'b1;
assign pslverr = 1'b0;


endmodule