
//cm3_fifo_demo
module cm3_fifo_demo
(
  sys_clk,
  reset_n,
  uart0_rxd,
  uart0_txd,
  TEST_WRITE,
  TEST_READ
);

input sys_clk;
input reset_n;

//UART0
input uart0_rxd;
output uart0_txd;

input TEST_WRITE;
input TEST_READ;


wire write_enable;
wire read_enable;
wire [31:0] user_addr;
wire [31:0] user_wdata;
wire [31:0] user_rdata;

wire test_wr;
wire test_rd;

//APB2 master bus
wire        master_pclk;
wire        master_prst;
wire        master_penable;
wire [7:0]  master_paddr;
wire        master_pwrite;
wire [31:0] master_pwdata;
wire [3:0]  master_pstrb;
wire [2:0]  master_pprot;
wire        master_psel1;
wire [31:0] master_prdata1;
wire        master_pready1;
wire        master_pslverr1;

//key_debounce instantiation
key_debounce u_rst
(
    .clk      (master_pclk),
    .rst_n    (1'b1),
    .key_1    (test_wr),
    .key1_n   (TEST_WRITE)
); 

//key_debounce instantiation
key_debounce u_rst_arb
(
    .clk      (master_pclk),
    .rst_n    (1'b1),
    .key_1    (test_rd),
    .key1_n   (TEST_READ)
);  

wire wr;
wire rd;

reg test_wr_d1;
reg test_rd_d1;

always@(posedge master_pclk or negedge master_prst)
begin
    if(!master_prst)
    begin
        test_wr_d1 <= 1'b0;
        test_rd_d1 <= 1'b0;
    end
    else
    begin
        test_wr_d1 <= test_wr;
        test_rd_d1 <= test_rd;
    end
end

assign wr = !test_wr & test_wr_d1;
assign rd = !test_rd & test_rd_d1;

//test_mode instantiation
test_mode u_test_mode
(
    .clk(master_pclk),
    .reset(master_prst),
    .test_write(wr),
    .test_read(rd),
    .write_enable(write_enable),
    .read_enable(read_enable),
    .user_addr(user_addr),
    .user_wdata(user_wdata),
    .user_rdata(user_rdata)
);

//apb_fifo instantiation
apb_fifo u_apb_fifo
(
    .pclk(master_pclk),
    .presetn(master_prst),
    .psel(master_psel1),
    .penable(master_penable),
    .pwrite(master_pwrite),
    .paddr({4'b0000,master_paddr[7:2]}),
    .pwdata(master_pwdata),
    .prdata(master_prdata1),
    .pready(master_pready1),
    .pslverr(master_pslverr1),
    .write_en(write_enable),
    .read_en(read_enable),
    .user_addr(user_addr),
    .user_wdata(user_wdata),
    .user_rdata(user_rdata)
);

//Gowin_EMPU_Top instantiation
Gowin_EMPU_Top u_Gowin_EMPU_Top
(
  .sys_clk(sys_clk),
  .reset_n(reset_n),
  .uart0_rxd(uart0_rxd),
  .uart0_txd(uart0_txd),
  .master_pclk(master_pclk),
  .master_prst(master_prst),
  .master_penable(master_penable),
  .master_paddr(master_paddr),
  .master_pwrite(master_pwrite),
  .master_pwdata(master_pwdata),
  .master_pstrb(master_pstrb),
  .master_pprot(master_pprot),
  .master_psel1(master_psel1),
  .master_prdata1(master_prdata1),
  .master_pready1(master_pready1),
  .master_pslverr1(master_pslverr1)
);

endmodule