/*
 ******************************************************************************************
 * @file      user_intr_demo.c
 * @author    GowinSemiconductor
 * @device    Gowin_EMPU(GW1NS-4C)
 * @brief     Demo User interrupt.
 ******************************************************************************************
 */

/* Includes ------------------------------------------------------------------*/
#include "demo.h"


/* Definitions ---------------------------------------------------------------*/

#if USER_INTR_DEMO

/* Includes: */
#include "uart.h"
#include "gpio.h"
#include "delay.h"
#include <stdio.h>


/* Definitions: */

//Application entry function
int user_intr_demo(void)
{
  SystemInit();       //Initializes system clock
  gpio_init();        //Initializes GPIO
  delay_init();       //Initializes delay functions
  uart_init(UART0,    //Initializes UART0
	        38400,    //Baudrate
	        1,        //Tx
	        1,        //Rx
	        0,        //Tx interrupt
	        1,        //Rx interrupt
	        0,        //Tx overflow interrupt
	        0);       //Rx overflow interrupt

  printf("Initializes system...\r\n");
  printf("Initializes GPIO...\r\n");
  printf("Initializes UART0...\r\n");
  printf("Initializes delay functions...\r\n");

  //Enable user interrupt 0
  nvic_irq_enable(USER_INT0_IRQn, 0, 0);
  printf("Enable user interrupt 0...\r\n");

  while(1)
  {
    GPIO_ResetBit(GPIO0,GPIO_Pin_0);	//LED1 on
    delay_sec(1);                    	//1 sec

    GPIO_SetBit(GPIO0,GPIO_Pin_0);		//LED1 off
    delay_sec(1);
  }
}

//User interrupt 0 handler function
void USER_INT0_Handler(void)
{
  printf("User interrupt 0 is triggered...\r\n");
}

#endif
